from libifstate.log import logger, IfStateLogging
from pyroute2 import IPRoute

from pyroute2.netlink.rtnl.tcmsg import tcmsg
from pyroute2.netlink.rtnl import RTM_DELTFILTER
from pyroute2.netlink import NLM_F_REQUEST
from pyroute2.netlink import NLM_F_ACK
from pyroute2.netlink import NLM_F_CREATE
from pyroute2.netlink import NLM_F_EXCL

try:
    # pyroute2 <0.6
    from pyroute2.ethtool.ioctl import SIOCETHTOOL
except ModuleNotFoundError:
    # pyroute2 >= 0.6
    from pr2modules.ethtool.ioctl import SIOCETHTOOL

import socket
import fcntl
import struct
import array
import struct

# ethtool helper
ETHTOOL_GDRVINFO = 0x00000003  # Get driver info
STRUCT_DRVINFO = struct.Struct(
    "I" +    # cmd
    "32s" +  # driver
    "32s" +  # version
    "32s" +  # fw_version
    "32s" +  # bus_info
    "32s" +  # reserved1
    "12s" +  # reserved2
    "I" +    # n_priv_flags
    "I" +    # n_stats
    "I" +    # testinfo_len
    "I" +    # eedump_len
    "I"      # regdump_len
)

ETHTOOL_GPERMADDR = 0x00000020  # Get permanent hardware address
L2_ADDRLENGTH = 6  # L2 address length
sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)


class IPRouteExt(IPRoute):
    def del_filter_by_info(self, index=0, handle=0, info=0, parent=0):
        msg = tcmsg()
        msg['index'] = index
        msg['handle'] = handle
        msg['info'] = info
        if parent != 0:
            msg['parent'] = parent

        return tuple(ipr.nlm_request(
            msg,
            msg_type=RTM_DELTFILTER,
            msg_flags=NLM_F_REQUEST |
            NLM_F_ACK | NLM_F_CREATE | NLM_F_EXCL
        ))

    def get_businfo(self, ifname):
        data = array.array("B", struct.pack(
            "I", ETHTOOL_GDRVINFO))
        data.extend(b'\x00' * (STRUCT_DRVINFO.size - len(data)))

        ifr = struct.pack('16sP', ifname.encode(
            "utf-8"), data.buffer_info()[0])

        try:
            r = fcntl.ioctl(sock.fileno(), SIOCETHTOOL, ifr)
        except OSError:
            return None

        drvinfo = STRUCT_DRVINFO.unpack(data)

        return drvinfo[4].decode('ascii').split('\x00')[0]

    def get_permaddr(self, ifname):
        data = array.array("B", struct.pack(
            "II", ETHTOOL_GPERMADDR, L2_ADDRLENGTH))
        data.extend(b'\x00' * L2_ADDRLENGTH)

        ifr = struct.pack('16sP', ifname.encode(
            "utf-8"), data.buffer_info()[0])

        try:
            r = fcntl.ioctl(sock.fileno(), SIOCETHTOOL, ifr)
        except OSError:
            return None

        l2addr = ":".join(format(x, "02x") for x in data[8:])
        if l2addr == "00:00:00:00:00:00":
            return None

        return l2addr

    def get_iface_by_businfo(self, businfo):
        for iface in iter(self.get_links()):
            ifname = iface.get_attr('IFLA_IFNAME')
            bi = self.get_businfo(ifname)

            if bi and bi == businfo:
                return iface['index']

    def get_iface_by_permaddr(self, permaddr):
        for iface in iter(self.get_links()):
            ifname = iface.get_attr('IFLA_IFNAME')
            addr = self.get_permaddr(ifname)

            if addr and addr == permaddr:
                return iface['index']

        return None

    def get_ifname_by_index(self, index):
        link = next(iter(ipr.get_links(index)), None)

        if link is None:
            return index

        return link.get_attr('IFLA_IFNAME')


ipr = IPRouteExt()



from ctypes import cdll, Structure, POINTER, c_int, c_uint, c_char_p, c_size_t, c_bool, c_void_p, c_ubyte, c_ulong, c_char
import ctypes


BPF_PROG_TYPE_XDP = 6


class struct_bpf_link(Structure):
    pass

class struct_bpf_object(Structure):
    pass

class struct_bpf_object_open_opts(Structure):
    _fields_ = [
    ('sz', c_size_t),
    ('object_name', c_char_p),
    ('relaxed_maps', c_bool),
    ('pin_root_path', c_char_p),
    ('unnamed_1', c_uint),
    ('kconfig', c_char_p),
    ('btf_custom_path', c_char_p),
    ('kernel_log_buf', c_char_p),
    ('kernel_log_size', c_size_t),
    ('kernel_log_level', c_uint),
    ('unnamed_2', c_size_t),
]

class struct_bpf_prog_info(Structure):
    _fields_ = [
    ('type', c_uint),
    ('id', c_uint),
    ('tag', c_ubyte * int(8)),
    ('jited_prog_len', c_uint),
    ('xlated_prog_len', c_uint),
    ('jited_prog_insns', c_ulong),
    ('xlated_prog_insns', c_ulong),
    ('load_time', c_ulong),
    ('created_by_uid', c_uint),
    ('nr_map_ids', c_uint),
    ('map_ids', c_ulong),
    ('name', c_char * int(16)),
    ('ifindex', c_uint),
    ('gpl_compatible', c_uint, 1),
    ('unnamed_1', c_uint, 31),
    ('netns_dev', c_ulong),
    ('netns_ino', c_ulong),
    ('nr_jited_ksyms', c_uint),
    ('nr_jited_func_lens', c_uint),
    ('jited_ksyms', c_ulong),
    ('jited_func_lens', c_ulong),
    ('btf_id', c_uint),
    ('func_info_rec_size', c_uint),
    ('func_info', c_ulong),
    ('nr_func_info', c_uint),
    ('nr_line_info', c_uint),
    ('line_info', c_ulong),
    ('jited_line_info', c_ulong),
    ('nr_jited_line_info', c_uint),
    ('line_info_rec_size', c_uint),
    ('jited_line_info_rec_size', c_uint),
    ('nr_prog_tags', c_uint),
    ('prog_tags', c_ulong),
    ('run_time_ns', c_ulong),
    ('run_cnt', c_ulong),
    ('recursion_misses', c_ulong),
    ('verified_insns', c_uint),
]


class struct_bpf_program(Structure):
    pass


# check if libbpf.so.1 is available
libbpf = None
try:
    libbpf = cdll.LoadLibrary('libbpf.so.1')
except OSError:
    # ignore missing library
    raise ModuleNotFoundError("Failed to load library libbpf.so.1!")


libbpf.bpf_object__open_file.argtypes = (
    c_char_p,                               # const char *path
    POINTER(struct_bpf_object_open_opts),   # struct bpf_object_open_opts *opts
)
libbpf.bpf_object__open_file.restype = POINTER(struct_bpf_object)

libbpf.bpf_program__attach_xdp.argtypes = (
    POINTER(struct_bpf_program),    # const struct bpf_program *prog
    c_int,                          # int ifindex
)
libbpf.bpf_program__attach_xdp.restype = POINTER(struct_bpf_link)

libbpf.bpf_object__next_program.argtypes = (
    POINTER(struct_bpf_object),     # const struct bpf_object *obj
    POINTER(struct_bpf_program),    # struct bpf_program *prog
)
libbpf.bpf_object__next_program.restype = POINTER(struct_bpf_program)

libbpf.bpf_program__section_name.argtypes = (
    POINTER(struct_bpf_program),    # const struct bpf_program *prog
)
libbpf.bpf_program__section_name.restype = c_char_p

libbpf.bpf_object__load.argtypes = (
    POINTER(struct_bpf_object),     # const struct bpf_object *obj
)
libbpf.bpf_object__load.restype = c_int

libbpf.bpf_object__name.argtypes = (
    POINTER(struct_bpf_object),     # const struct bpf_object *obj
)
libbpf.bpf_object__name.restype = c_char_p


libbpf.bpf_link__pin.argtypes = (
    POINTER(struct_bpf_link),   # const struct bpf_link *link
    c_char_p,                   # const char *path
)
libbpf.bpf_link__pin.restype = c_int


libbpf.bpf_program__fd.argtypes = (
    POINTER(struct_bpf_program),    # const struct bpf_program *prog
)
libbpf.bpf_program__fd.restype = c_int


libbpf.bpf_prog_get_fd_by_id.argtypes = (
    c_uint,     # c_uint id
)
libbpf.bpf_prog_get_fd_by_id.restype = c_int


libbpf.bpf_obj_get_info_by_fd.argtypes = (
    c_int,      # int bpf_fd
    c_void_p,   # void *info
    POINTER(c_uint),   # c_uint *info_len
)
libbpf.bpf_obj_get_info_by_fd.restype = c_int



from libifstate.util import logger, ipr, IfStateLogging
from libifstate.exception import netlinkerror_classes
from libifstate.bpf import libbpf, struct_bpf_prog_info

from pyroute2.netlink.rtnl.ifinfmsg import XDP_FLAGS_SKB_MODE
from pyroute2.netlink.rtnl.ifinfmsg import XDP_FLAGS_DRV_MODE
from pyroute2.netlink.rtnl.ifinfmsg import XDP_FLAGS_HW_MODE

import ctypes
import os

class XDP():
    def __init__(self, iface, xdp):
        self.iface = iface
        self.xdp = xdp

    def apply(self, do_apply):
        self.link = next(iter(ipr.get_links(ifname=self.iface)), None)

        if self.link == None:
            logger.warning('link missing', extra={'iface': self.iface})
            return

        # get ifindex
        self.idx = self.link['index']

        # get current BPF tag, if any
        current_prog_tag = None
        current_prog_id = self.link.get_attr('IFLA_XDP', {}).get('IFLA_XDP_PROG_ID')
        current_attached = self.link.get_attr('IFLA_XDP', {}).get('IFLA_XDP_ATTACHED')
        if current_prog_id:
            current_prog_fd = libbpf.bpf_prog_get_fd_by_id(current_prog_id)
            if current_prog_fd < 0:
                logger.warning('could not get current XDP prog fd for {}: {}'.format(
                    self.iface, os.strerror(-current_prog_fd)))
            else:
                current_prog_info = struct_bpf_prog_info()
                current_prog_size = ctypes.c_uint(ctypes.sizeof(current_prog_info))
                rc = libbpf.bpf_obj_get_info_by_fd(current_prog_fd, ctypes.byref(current_prog_info), ctypes.byref(current_prog_size))
                if rc == 0:
                    current_prog_tag = bytes(current_prog_info.tag).hex()

                    logger.debug('current prog tag: {}'.format(current_prog_tag), extra={
                                'iface': self.iface})
                else:
                    logger.warning('could not get current XDP obj info for {}: {}'.format(
                        self.iface, os.strerror(-rc)))

        logger.debug('current attached: {}'.format(current_attached), extra={
            'iface': self.iface})

        # load new BPF prog
        new_prog_tag = None
        new_prog_fd = None
        if not self.xdp:
            new_prog_fd = -1
        elif 'pinned' in self.xdp:
            fh = open(self.xdp["pinned"], 'r')
            new_prog_fd = fh.fileno()
        elif 'object' in self.xdp:
            obj = libbpf.bpf_object__open_file(
                os.fsencode(self.xdp["object"]),
                None,
            )
            if not obj:
                logger.warning('XDP open on {} failed: {}'.format(
                    self.iface, os.strerror(-current_prog_fd)))
                return

            logger.debug('loaded object: {}'.format(libbpf.bpf_object__name(obj).decode('ascii')), extra={
                        'iface': self.iface})

            rc = libbpf.bpf_object__load(obj)
            if rc < 0:
                logger.warning('XDP load on {} failed: {}'.format(
                    self.iface, os.strerror(-rc)))
                return

            prog = libbpf.bpf_object__next_program(obj, None)
            while prog:
                section = libbpf.bpf_program__section_name(prog).decode('ascii')

                logger.debug('  section: {}'.format(section), extra={
                            'iface': self.iface})

                if section == self.xdp.get('section', 'xdp'):
                    break
                prog = libbpf.bpf_object__next_program(obj, prog)

            if not prog:
                logger.warning('XDP section {} on {} not found'.format(
                    self.xdp.get('section'), self.iface))
                return

            # get prog fd
            new_prog_fd = libbpf.bpf_program__fd(prog)
            if not new_prog_fd:
                logger.warning('XDP failed to get prog fd on {}'.format(
                    self.iface))
                return
        else:
            assert False, "unhandled XDP settings"

        # get new prog tag
        if new_prog_fd != -1:
            info = struct_bpf_prog_info()
            sz = ctypes.c_uint(ctypes.sizeof(info))
            rc = libbpf.bpf_obj_get_info_by_fd(new_prog_fd, ctypes.byref(info), ctypes.byref(sz))
            if rc == 0:
                new_prog_tag = bytes(info.tag).hex()

                logger.debug('new prog tag: {}'.format(new_prog_tag), extra={
                            'iface': self.iface})
            else:
                logger.warning('XDP failed to get prog info on {}'.format(
                    self.iface))

        # get attach mode flags
        new_attached = self.xdp.get("mode", "auto")
        if type(new_attached) != list:
            new_attached = [new_attached]

        new_flags = 0
        if "auto" in new_attached:
            new_attached = ["xdp", "xdpgeneric", "xdpoffload"]
        else:
            if "xdp" in new_attached:
                new_flags = new_flags | XDP_FLAGS_DRV_MODE
            if "xdpgeneric" in new_attached:
                new_flags = new_flags | XDP_FLAGS_SKB_MODE
            if "xdpoffload" in new_attached:
                new_flags = new_flags | XDP_FLAGS_HW_MODE

        logger.debug('new attached: {}'.format(", ".join(new_attached)), extra={
            'iface': self.iface})

        logger.debug('new flags: {}'.format(new_flags), extra={
            'iface': self.iface})

        # set new XDP prog if tag or attach mode has changed
        if current_prog_tag != new_prog_tag or not current_attached in new_attached:
            if do_apply:
                try:
                    ipr.link('set', index=self.idx, xdp_fd=new_prog_fd, xdp_flags=new_flags)
                except Exception as err:
                    if not isinstance(err, netlinkerror_classes):
                        raise

                    if new_prog_fd == -1:
                        logger.warning('attaching XDP program on {} failed: {}'.format(
                            self.iface, err.args[1]))
                    else:
                        # try again: detaching XDP first
                        try:
                            ipr.link('set', index=self.idx, xdp_fd=-1)

                            ipr.link('set', index=self.idx, xdp_fd=new_prog_fd, xdp_flags=new_flags)
                        except Exception as err:
                            if not isinstance(err, netlinkerror_classes):
                                raise

                            logger.warning('attaching XDP program on {} failed: {}'.format(
                                self.iface, err.args[1]))

            if new_prog_fd == -1:
                logger.info('detach', extra={
                            'iface': self.iface, 'style': IfStateLogging.STYLE_DEL})
            else:
                logger.info('change', extra={
                            'iface': self.iface, 'style': IfStateLogging.STYLE_CHG})
        else:
            logger.info(
                'ok',
                extra={'iface': self.iface, 'style': IfStateLogging.STYLE_OK})

try:
    libbpf.bpf_obj_get_info_by_fd
    libbpf.bpf_object__name
    libbpf.bpf_object__next_program
    libbpf.bpf_object__load
    libbpf.bpf_object__open_file
    libbpf.bpf_prog_get_fd_by_id
    libbpf.bpf_program__fd
    libbpf.bpf_program__section_name
except AttributeError as ex:
    # ignore missing library
    raise ModuleNotFoundError(ex)



from setuptools import setup, find_packages
from datetime import date


def readme():
    with open("README.md") as f:
        return f.read()


def version():
    with open("libifstate/__init__.py") as fd:
        for line in fd:
            if line.startswith('__version__'):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


def install_requires():
    requires = [
        "jsonschema"
    ]

    try:
        import pyroute2
    except ModuleNotFoundError:
        requires.append('pyroute2')

    return requires


setup(
    name="ifstate",
    version=version(),
    description="Manage host interface settings in a declarative manner",
    author="Thomas Liske",
    author_email="thomas@fiasko-nw.net",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://ifstate.net/",
    license="GPL3+",
    packages=find_packages(),
    package_data={
        "libifstate": ["../schema/*.schema.json"],
    },
    install_requires=install_requires(),
    extras_require={
        'wireguard': [
            'wgnlpy'
        ]
    },
    entry_points={
        "console_scripts": ["ifstatecli = ifstate.ifstate:main"]
    },
)



#!/usr/bin/env python3

from libifstate.parser import YamlParser
from libifstate import __version__, IfState
from libifstate.exception import FeatureMissingError, LinkNoConfigFound, ParserValidationError, ParserOpenError, ParserIncludeError
from libifstate.util import logger, IfStateLogging
from collections import namedtuple
from copy import deepcopy

import argparse
import logging
import re
import signal
import sys
import yaml


class Actions():
    CHECK = "check"
    APPLY = "apply"
    SHOW = "show"
    SHOWALL = "showall"
    VRRP = "vrrp"
    VRRP_FIFO = "vrrp-fifo"
    SHELL = "shell"

ACTIONS_HELP = {
    "CHECK"    : "dry run update the network config",
    "APPLY"    : "update the network config",
    "SHOW"     : "show running network config",
    "SHOWALL"  : "show running network config (more settings)",
    "VRRP"     : "run as keepalived notify script",
    "VRRP_FIFO": "run as keepalived notify_fifo_script",
    "SHELL"    : "launch interactive python shell (pyroute2)",
}

def shell():
    import readline # optional, will allow Up/Down/History in the console
    import code
    from pprint import pprint
    from libifstate.util import ipr

    print("""Symbols:
  ipr: pyroute2.IPRoute() object
  pprint: Data pretty printer function
""")

    print("Interfaces:")
    for link in ipr.get_links():
        print("  {:2}: {}".format(link.get('index'),link.get_attr( 'IFLA_IFNAME')))
    print("")

    variables = {
        'ipr': ipr,
        'pprint': pprint,
    }

    shell = code.InteractiveConsole(variables)
    shell.interact()


def main():
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group()
    parser.add_argument('--version', action='version',
                        version='%(prog)s {version}'.format(version=__version__))
    group.add_argument("-v", "--verbose", action="store_true",
                       help="be more verbose")
    group.add_argument("-q", "--quiet", action="store_true",
                       help="be more quiet, print only warnings and errors")
    parser.add_argument("-s", "--soft-schema", action="store_true",
                        help="ignore schema validation errors, expect ifstatecli to trigger internal exceptions")
    parser.add_argument("-c", "--config", type=str,
                        default="/etc/ifstate/config.yml", help="configuration YaML filename")
    subparsers = parser.add_subparsers(
        dest='action', required=True, help="specifies the action to perform")

    action_parsers = {
        a.lower().replace("_", "-"): subparsers.add_parser(a.lower().replace("_", "-"), help=ACTIONS_HELP[a]) for a in dir(Actions) if not a.startswith('_')
    }

    # Parameters for the vrrp action
    action_parsers[Actions.VRRP].add_argument(
        "type", type=str.lower, choices=["group", "instance"], help="type of vrrp notification")
    action_parsers[Actions.VRRP].add_argument(
        "name", type=str, help="name of the vrrp group or instance")
    action_parsers[Actions.VRRP].add_argument(
        "state", type=str.lower, choices=["unknown", "fault", "backup", "master"], help="the new state for the vrrp group or instance")

    # Parameters for the vrrp-fifo action
    action_parsers[Actions.VRRP_FIFO].add_argument(
        "fifo", type=str, help="named FIFO to read state changes from")

    args = parser.parse_args()
    if args.verbose:
        lvl = logging.DEBUG
    elif args.quiet:
        lvl = logging.ERROR
    else:
        lvl = logging.INFO

    if args.action == Actions.SHELL:
        shell()
        exit(0)

    ifslog = IfStateLogging(lvl)
    ifs = IfState()

    if args.action in [Actions.SHOW, Actions.SHOWALL]:
        # preserve dict order on python 3.7+
        if sys.version_info >= (3, 7):
            yaml.add_representer(
                dict, lambda self, data: yaml.representer.SafeRepresenter.represent_dict(self, data.items()))
        print(yaml.dump(ifs.show(args.action == Actions.SHOWALL)))

        ifslog.quit()
        exit(0)

    if args.action in [Actions.CHECK, Actions.APPLY, Actions.VRRP, Actions.VRRP_FIFO]:
        try:
            parser = YamlParser(args.config)
        except ParserOpenError as ex:
            logger.error(
                "Config loading from {} failed: {}".format(ex.fn, ex.msg))
            ifslog.quit()
            exit(1)
        except yaml.parser.ParserError as ex:
            logger.error("Config parsing failed:\n\n{}".format(str(ex)))
            ifslog.quit()
            exit(2)
        except ParserIncludeError as ex:
            logger.error(
                "Config include file {} failed: {}".format(ex.fn, ex.msg))
            ifslog.quit()
            exit(3)

        try:
            ifstates = parser.config()
            ifs.update(ifstates, args.soft_schema)
        except ParserValidationError as ex:
            logger.error("Config validation failed for {}".format(ex.detail))
            ifslog.quit()
            exit(4)
        except FeatureMissingError as ex:
            logger.error(
                "Config uses unavailable feature: {}".format(ex.feature))
            ifslog.quit()
            exit(5)

        if args.action == Actions.CHECK:
            try:
                ifs.check()
            except LinkNoConfigFound:
                pass
        elif args.action == Actions.VRRP_FIFO:
            status_pattern = re.compile(
                r'(group|instance) "([^"]+)" (unknown|fault|backup|master)$', re.IGNORECASE)

            with open(args.fifo) as fifo:
                for line in fifo:
                    m = status_pattern.match(line.strip())
                    if m:
                        signal.signal(signal.SIGHUP, signal.SIG_IGN)
                        signal.signal(signal.SIGPIPE, signal.SIG_IGN)
                        signal.signal(signal.SIGTERM, signal.SIG_IGN)
                        try:
                            ifs_tmp = deepcopy(ifs)
                            ifs_tmp.apply(m.group(1), m.group(2), m.group(3))
                        except:
                            pass
        else:
            # ignore some well-known signals to prevent interruptions (i.e. due to ssh connection loss)
            signal.signal(signal.SIGHUP, signal.SIG_IGN)
            signal.signal(signal.SIGPIPE, signal.SIG_IGN)
            signal.signal(signal.SIGTERM, signal.SIG_IGN)
            try:
                if args.action == Actions.APPLY:
                    ifs.apply()
                elif args.action == Actions.VRRP:
                    ifs.apply(
                        args.type, args.name, args.state)
            except LinkNoConfigFound:
                pass

        ifslog.quit()


if __name__ == "__main__":
    main()



from libifstate.util import logger, ipr, IfStateLogging
from libifstate.exception import netlinkerror_classes


class BRPort():
    def __init__(self, iface, brport):
        self.iface = iface
        self.brport = brport

    def depends(self):
        '''get link dependencies (i.e. backup port)'''

        backup_port = self.brport.get('backup_port')

        if backup_port:
            return (backup_port,)
        else:
            return ()

    def has_changes(self, idx):
        logger.debug('checking brport', extra={'iface': self.iface})

        has_changes = False
        brport_state = next(iter(ipr.brport('dump', index=idx)), None)

        # no a bridge port
        if brport_state is None:
            logger.debug('  not a bridge port', extra={'iface': self.iface})
            return True

        for setting in self.brport.keys():

            current_value = next(iter(next(iter(brport_state.get_attrs(
                'IFLA_PROTINFO'))).get_attrs('IFLA_BRPORT_{}'.format(setting.upper()))))

            logger.debug('  %s: %s => %s', setting, current_value, self.brport[setting], extra={
                         'iface': self.iface})
            has_changes |= current_value != self.brport[setting]

        return has_changes

    def apply(self, do_apply, idx, excpts):
        brport_state = ipr.brport('dump', index=idx)
        if brport_state == None:
            logger.warning('link is not a bridge port',
                           extra={'iface': self.iface})
            return

        logger.info(
            'change (brport)', extra={'iface': self.iface, 'style': IfStateLogging.STYLE_CHG})

        logger.debug("bridge link set: {}".format(
            " ".join("{}={}".format(k, v) for k, v in self.brport.items())))

        if do_apply:
            try:
                ipr.brport('set', index=idx, **(self.brport))
            except Exception as err:
                if not isinstance(err, netlinkerror_classes):
                    raise
                excpts.add('brport', err, **(self.brport))
